package ima.uco.fr.nuit.model;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import ilog.concert.IloException;
import ilog.concert.IloIntVar;
import ilog.concert.IloLinearIntExpr;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplex.UnknownObjectException;

/**
 * Modle mathmatique CPLEX
 * @author IMAvengers
 *
 */
public class Model {

	/**
	 * Dimension du panneau publicitaire
	 */
	private int nbLigne, nbColonne;

	/**
	 * Constantes lis aux dimensions du panneau
	 */
	private int K; 				// Nombre totales de diagonales
	private int R;				// Nombre de femmes maximum possibles par colonnes ajout  1
	private int[] pourcentage;	// Pourcentages de nombre de femmes par diagonales pouvant tre obtenus

	private IloCplex model; 	// Modle Cplex

	/**
	 * VARIABLES DE DECISIONS
	 */
	private IloIntVar[][] x; 	// x[i][j] : 1 si la personne de la ligne i, en colonne j est une Femme, 0 sinon
	private IloIntVar[] alpha; 	// alpha[t] : 1 si il y a t femmes sur chaque ligne, 0 sinon
	private IloIntVar[] y; 		// y[j] : nombre de femmes sur la colonne j
	private IloNumVar[] c; 		// c[k] : pourcentage de femmes sur la diagonale k
	private IloIntVar[][] z; 	// z[k][l] : 1 si c[k]=l, 0 sinon   
	private IloIntVar[] u; 		// u[l] : 1 si l est un % existant sur au moins une diagonale, 0 sinon
	private IloIntVar[][] v; 	// v[j][r] : 1 si y[j]=R, 0 sinon
	private IloIntVar p;		// Variable de travail reprsentant le nombre final de femmes par ligne 

	/**
	 * Constructeur 
	 * @param nbLigne : nombre de ligne dans le panneau publicitaire
	 * @param nbColonne : nombre de colonne dans le panneau publicitaire
	 */
	public Model(int nbLigne, int nbColonne, int[]pourcentage) {
		// Initialisation des donnes
		this.nbLigne = nbLigne;
		this.nbColonne = nbColonne;
		this.pourcentage = pourcentage;
		this.R = nbLigne+1;
		this.K = nbLigne*4-2;

		try {
			// Cration du modle CPLEX
			model= new IloCplex();

			buildVariables();	// Construction des variables des dcisions
			buildConstraints(); // Construction des contraintes 
			buildObjective();	// Construction de la fonction objectif

		} catch (IloException e) {
			e.printStackTrace();
		}
	}

	/**
	 * variable's construction
	 * 
	 * @throws IloException
	 */
	private void buildVariables() throws IloException {
		// Initialisation de x_ij
		x = new IloIntVar[nbLigne][nbColonne];
		for (int i=0; i<nbLigne; i++) 
			for(int j=0; j <nbColonne; j++)
				x[i][j] = model.boolVar("x_"+(i+1)+"_"+(j+1));

		// Initialisation de y_j
		y = new IloIntVar[nbColonne];
		for(int j=0; j <nbColonne; j++)
			y[j] = model.intVar(0, nbLigne, "y_"+(j+1));

		// Initialisation de c_k
		c = new IloNumVar[K];		
		for(int k=0; k<K; k++)
			c[k] = model.numVar(0, 100, "c_"+(k+1));

		// Initialisation de alpha_t
		alpha = new IloIntVar[nbColonne+1];
		for (int t=0; t<=nbColonne; t++)
			alpha[t] = model.boolVar("alpha_"+t); // t varie entre 0 et nombre de colonne dans la grille

		// Initialisation de z_kl
		z = new IloIntVar[K][pourcentage.length];
		for(int k=0; k<K; k++)
			for(int l=0; l<pourcentage.length; l++)
				z[k][l] = model.boolVar("z_"+(k+1)+"_"+(l+1));

		// Initialisation de u_l
		u = new IloIntVar[pourcentage.length];
		for(int l=0; l<pourcentage.length; l++)
			u[l] = model.boolVar("u_"+(l+1));

		// Initialisation de v_jr
		v = new IloIntVar[nbColonne][R];
		for (int j=0; j<nbColonne; j++) 
			for (int r=0; r<R; r++)
				v[j][r]= model.boolVar("v_"+(j+1)+"_"+r); 

		// Initialisation de p
		p = model.intVar(0, nbLigne, "p");
	}

	/**
	 * Construction of the objectiv function
	 * 
	 * @throws IloException
	 */
	private void buildObjective() throws IloException {
		// Initialisation
		IloLinearIntExpr of = model.linearIntExpr();

		for(int l=0; l<pourcentage.length; l++)
			of.addTerm(1000, u[l]);

		//Ajout de la fonction objectif au model
		model.addMaximize(of, "OF");
	}

	/**
	 * Constraints' construction
	 * @throws IloException
	 */
	private void buildConstraints() throws IloException {
		// Contrainte n1 : pourcentage de femmmes le mme sur toutes les lignes
		for(int i=0; i<nbLigne; i++)
			// (1) Pour toutes les lignes, somme sur j des x_ij = p
			model.addEq(model.sum(x[i]), p, "C1_"+i);

		// (2) Somme sur t des t.alpha_t = p 
		IloLinearNumExpr c1 = model.linearNumExpr();
		for(int t=0; t<=nbLigne; t++)
			c1.addTerm(t, alpha[t]);
		model.addEq(c1, p, "C2");

		// (3) Somme sur t des alpha_t = 1
		model.addEq(model.sum(alpha), 1, "C3");

		// Contrainte n2 : Chaque colonne doit contenir un pourcentage de femmes diffrent
		for (int j=0; j<nbColonne; j++) {
			// (4) Pour toutes les colonnes, somme sur i des x_ij = y_j
			IloLinearNumExpr c2 = model.linearNumExpr();
			for(int i=0; i<nbLigne; i++)
				c2.addTerm(1, x[i][j]);
			model.addEq(c2, y[j], "C4_"+j);

			// (5) Pour toutes les colonnes, y_j = somme sur r de r.v_jr
			IloLinearNumExpr c3 = model.linearNumExpr();
			for(int r=0; r<R; r++)
				c3.addTerm(r, v[j][r]);
			model.addEq(y[j], c3, "C5_"+j);

			// (6) Pour toutes les colonnes, somme sur r de v_jr = 1
			model.addEq(model.sum(v[j]), 1, "C6_"+j);
		}

		// (7) Pour toutes les colonnes, somme sur j de v_jr <= 1
		for(int r=0; r<R; r++){
			IloLinearNumExpr c4 = model.linearNumExpr();
			for (int j=0; j<nbColonne; j++)
				c4.addTerm(v[j][r], 1);
			model.addLe(c4, 1, "C7_"+r);
		}

		// Contrainte n3 : Nombre de publicit (diagonale)
		for(int k=1; k<K+1; k++){
			// (8) Calcul selon les colonnes
			IloLinearNumExpr cq = model.linearNumExpr();
			double temp = 0;
			if (k < nbLigne+1) {
				temp = k;
				for (int i=1; i<k+1; i++) 
					cq.addTerm(1, x[k-i+1-1][i-1]);

			} else if (k < 2*nbLigne) {
				temp = 2*nbLigne - k;
				for (int i=k-nbLigne; i<nbLigne; i++) 
					cq.addTerm(1, x[k-i-1][i+1-1]);

			} else if (k < 3*nbLigne) {
				temp = k - (2*nbLigne-1);
				for (int i=1; i<k-2*nbLigne+2; i++) 
					cq.addTerm(1, x[i-1][i-k+3*nbLigne-1-1]);

			} else {
				temp = (4*nbLigne - 1) - k;
				for (int i=k-3*nbLigne+2; i<nbLigne+1; i++) 
					cq.addTerm(1, x[i-1][3*nbLigne-1-k+i-1]);
			}

			model.addEq(model.prod(100, cq), model.prod(temp, c[k-1]), "CQ_"+k); 
		}

		for(int k=0; k<K; k++){
			// (9) Pour toutes les colonnes, Multiplicateur. c_k = somme sur l de l.z_kl
			IloLinearNumExpr c5 = model.linearNumExpr();
			for(int l=0; l<pourcentage.length;l++){
				c5.addTerm(pourcentage[l], z[k][l]);
			}
			model.addEq(model.prod(Main.calculeMultiplicateur(), c[k]), c5, "C8_"+k);

			// (10) Pour toutes les colonnes, somme sur l des z_kl = 1
			model.addEq(model.sum(z[k]), 1, "C9_"+k);	
		}

		for(int l=0; l<pourcentage.length;l++){
			IloLinearNumExpr c6= model.linearNumExpr();
			for(int k=0; k<K; k++)
				c6.addTerm(z[k][l], 1);
			
			// (11) Pour tous les diffrents pourcentages des diagonales, somme sur k des z_kl >= u_l
			model.addGe(c6, u[l], "C10_"+l);

			// (12) Pour tous les diffrents pourcentages des diagonales, K.u_l >= somme sur k des z_kl
			model.addGe(model.prod(K, u[l]), c6, "C11_"+l);
		}

	}

	/**
	 * Rsout le modle mathmatique
	 * @return true s'il existe une solution, false sinon
	 */
	public boolean resolve(){
		try {
			model.setParam(IloCplex.IntParam.MIPEmphasis, 1);

			if (model.solve()) {
				// Affichage console des rsultats
				affichConsol();
			} 
			else 
				System.err.println("modle non rsolu");

			return true;

		} catch (IloException e) {
			e.printStackTrace();
		}

		return false;
	}

	/**
	 * Affiche dans la console les rsultats
	 * @throws UnknownObjectException
	 * @throws IloException
	 */
	private void affichConsol() throws UnknownObjectException, IloException {
		// Affiche la matrice x_ij
		System.out.println("Matrice x_ij :");
		for(int i=0;i<nbLigne;i++){
			for(int j=0;j<nbColonne;j++)
				System.out.print(model.getValue(x[i][j])+"\t");
			System.out.println();
		}
		System.out.println("\n");
		
		// Affiche la matrice y_j
		System.out.println("Matrice y_j :");	
		for(int j=0;j<nbColonne;j++)
			System.out.print(model.getValue(y[j])+"\t");
		System.out.println("\n");
		
		// Affiche la matrice c_k
		System.out.println("Matrice c_k :");
		for(int k=0; k<K; k++)
			System.out.print(model.getValue(c[k])+"\t");
		System.out.println("\n");
		
		// Affiche la matrice alpha_t
		System.out.println("Matrice alpha_t :");
		for (int t=0; t<=nbColonne; t++)
			System.out.print(model.getValue(alpha[t])+"\t");
		System.out.println("\n");
		
		// Affiche la matrice z_kl
		System.out.println("Matrice z_kl :");
		for(int k=0; k<K; k++)
			for(int l=0; l<pourcentage.length; l++)
				System.out.print(model.getValue(z[k][l])+"\t");
		System.out.println("\n");
		
		// Affiche la matrice u_l
		System.out.println("Matrice u_l :");
		for(int l=0; l<pourcentage.length; l++)
			System.out.print(model.getValue(u[l])+"\t");
		System.out.println("\n");
		
		// Affiche la matrice v_jr
		System.out.println("Matrice v_jr :");
		for (int j=0; j<nbColonne; j++) 
			for (int r=0; r<R; r++)
				System.out.print(v[j][r]);
		System.out.println("\n");

		// Affiche la valeur de p
		System.out.println("Le nombre de femmes par ligne est : " + model.getValue(p));
		
	}

	/**
	 * Cr le fichier LP du modle
	 */
	public void exportModelLP() {
		try {
			model.exportModel("model.lp");

		} catch (IloException e) {
			e.printStackTrace();
		}	
	}

	/**
	 * Export des solutions sous fichier CSV
	 * @param dossierOut 
	 */
	public void exportToCSV(String dossierOut) {
		try {
			String file = dossierOut.concat("solution_" + nbLigne + ".csv");
			PrintWriter out = new PrintWriter(new FileWriter(file ));
			
			for(int i=0;i<nbLigne;i++){
				for(int j=0;j<nbColonne;j++)
					out.print((int) model.getValue(x[i][j])+";");
				out.println();
			}
			
			out.close();

		} catch (IOException e) {
			e.printStackTrace();
		} catch (UnknownObjectException e) {
			e.printStackTrace();
		} catch (IloException e) {
			e.printStackTrace();
		}	
	}

	/**
	 * Transforme les rsultats CPLEX en forme de matrice
	 * @return la matrice x
	 */
	public int[][] obtainMatrix() {
		int[][] tab_x = new int[nbLigne][nbColonne];

		try {
			for(int i=0;i<nbLigne;i++)
				for(int j=0;j<nbColonne;j++)
					tab_x[i][j] = (int) model.getValue(x[i][j]);

		} catch (UnknownObjectException e) {
			e.printStackTrace();
		} catch (IloException e) {
			e.printStackTrace();
		}
		
		return tab_x;
	}
}